import numpy as np
from .signal import *


def test_vdot():
    assert signal([1, 2j], start=-1).vdot(signal([3, 4])) == -6j


def test_add():
    a = signal([1, 2j], start=-1) + signal([3, 4])
    assert a.isclose(signal([1, 3 + 2j, 4], start=-1))


def test_sub():
    a = signal([1, 2j], start=-1) - signal([3, 4])
    assert a.isclose(signal([1, -3 + 2j, -4], start=-1))


def test_modulate():
    a = signal([-3, -2, -1, 0, 1, 2, 3, 4], start=-3)
    A = signal([3, -2, 1, 0, -1, 2, -3, 4], start=-3)
    assert a.modulate(-1 + 0j).isclose(A)
    assert a.modulate(-1.).isclose(A)


def test_reverse():
    N = 100
    a = signal(np.random.rand(N), np.random.randint(-N, N))
    a_rev = a.reverse()

    assert a_rev.reverse().isclose(a)
    assert all(a[n] == a_rev[-n] for n in a.range)
    assert all(a_rev[n] == a[-n] for n in a_rev.range)


def test_downsample():
    a = signal([0, 1, 2, 3, 4, 5])
    assert a.downsample().isclose(signal([0, 2, 4]))

    a = signal([-3, -2, -1, 0, 1, 2], start=-3)
    assert a.downsample().isclose(signal([-2, 0, 2], start=-1))

    N = 100
    a = signal(np.random.rand(N), np.random.randint(-N, N))
    b = a.downsample()
    for n in a.range:
        if n % 2 == 0:
            assert b[n // 2] == a[n]


def test_upsample():
    a = signal([0, 1, 2, 3, 4, 5])
    assert a.upsample().isclose(signal([0, 0, 1, 0, 2, 0, 3, 0, 4, 0, 5]))

    a = signal([-3, -2, -1, 0, 1, 2], start=-3)
    assert a.upsample().isclose(
        signal([-3, 0, -2, 0, -1, 0, 0, 0, 1, 0, 2], start=-6))

    N = 100
    a = signal(np.random.rand(N), np.random.randint(-N, N))
    b = a.upsample()
    for n in b.range:
        assert b[n] == (a[n // 2] if n % 2 == 0 else 0)



import numpy as np
from .utils import *
from .signal import *
from .wavelets import *
from .hilbert import *

__all__ = ['mera1d', 'mera2d']


class mera1d:
    """1D Gaussian MERA for approximate ground state of free-fermion Hamiltonian at half filling."""

    def __init__(self, h, g):
        """The wavelet instances h, g should form an approximate Hilbert pair."""
        self.h, self.g = h, g

    @staticmethod
    def selesnick(K, L):
        return mera1d(*selesnick_hwlet(K, L))

    def eigenmode_pair(self, level, x=0):
        """
        Return approximate (negative-energy) eigenmode pair (a,b) on even/odd sublattices that arise from inserting
        unit signals into the given level of the inverse wavelet transforms (level=1, 2, ...).
        """
        assert level >= 1
        a = self.h.reconstruct(wavelet=signal([1], start=x))
        b = self.g.reconstruct(wavelet=signal([1], start=x))
        for _ in range(level - 1):
            a = self.h.reconstruct(scaling=a)
            b = self.g.reconstruct(scaling=b)
        return a, b

    def eigenmode(self, level, x=0, positive_energy=False):
        """
        Return approximate (negative-energy) eigenmode on original lattice that arises from the given level of the MERA
        (level=1, 2, ...).
        """
        assert level >= 1
        a, b = self.eigenmode_pair(level, x)
        a = a.modulate(-1.).upsample()
        b = b.modulate(-1.).upsample().shift(1)
        if not positive_energy:
            psi = (a + b) / np.sqrt(2)
        else:
            psi = (a - b) / np.sqrt(2)
        return psi

    @staticmethod
    def energy_of_mode(psi):
        """Compute energy of given single-particle mode."""
        return -2 * np.real(psi.vdot(psi.shift(-1)))

    def energy(self, levels):
        """Compute energy of approximate ground state with levels MERA layers."""
        E = []
        for level in range(1, levels + 1):
            psi = self.eigenmode(level)
            E.append(mera1d.energy_of_mode(psi) / 2**(level + 1))
        return np.sum(E)

    def correlation(self, dx, levels, x=None):
        """Compute correlation function C(x, x+dx) of approximate ground state with levels MERA layers."""
        if x is None:
            x = np.array([0])

        C = np.zeros(shape=(x.size, dx.size))
        for level in range(1, levels + 1):
            psi = self.eigenmode(level)
            for i, the_x in enumerate(x):
                for j, the_dx in enumerate(dx):
                    the_y = the_x + the_dx
                    C[i, j] += psi.shift(-the_y).downsample(level + 1).vdot(
                        psi.shift(-the_x).downsample(level + 1))

        return C

    def covariance(self, stop, levels, start=None):
        """Return covariance matrix <a_i^\dagger a_j> of subsystem {start,...,stop-1}."""
        if start is None:
            start = 0
        x = np.arange(start, stop)
        C = np.zeros(shape=(x.size, x.size))
        for level in range(1, levels + 1):
            psi = self.eigenmode(level)
            for i, the_x in enumerate(x):
                for j, the_y in enumerate(x):
                    C[i, j] += psi.shift(-the_y).downsample(level + 1).vdot(
                        psi.shift(-the_x).downsample(level + 1))
        return C

    def h_scaling(self, level, k):
        """
        Return single-particle scaling Hamiltonian (renormalized Hamiltonian) at given level
        (level = 0 is the original Hamiltonian).
        """
        h_10 = self._h_renormalized(level, k)
        return np.array([
            [np.zeros_like(k), h_10.conj()],
            [h_10, np.zeros_like(k)],
        ])

    def e_scaling(self, level, k):
        """
        Return dispersion relation of scaling Hamiltonian (renormalized Hamiltonian) at given level
        (level 0 is the original Hamiltonian).
        """
        return np.abs(self._h_renormalized(level, k))

    def e_wavelet(self, level, k):
        """Return dispersion relation of level-l wavelet Hamiltonian."""
        assert level >= 1
        return np.abs(self._h_renormalized(level, k, wavelet=True))

    def _h_renormalized(self, level, k, wavelet=False):
        """Return [1,0] matrix element of level-l single-particle Hamiltonian in k-space."""
        if level == 0:
            return np.exp(1j * k) - 1

        H = self.h.wavelet_filter.ft if wavelet else self.h.scaling_filter.ft
        G = self.g.wavelet_filter.ft if wavelet else self.g.scaling_filter.ft
        a = G(k / 2).conj() * H(k / 2) * self._h_renormalized(level - 1, k / 2)
        b = G(k / 2 + np.pi).conj() * H(k / 2 + np.pi) * self._h_renormalized(
            level - 1, k / 2 + np.pi)
        return (a + b) / 2


class mera2d:
    """2D Gaussian MERA for approximate ground state of free-fermion Hamiltonian at half filling."""

    def __init__(self, h, g):
        """The wavelet instances h, g should form an approximate Hilbert pair."""
        self.mera1d = mera1d(h, g)

    @staticmethod
    def selesnick(K, L):
        return mera2d(*selesnick_hwlet(K, L))

    def eigenmode_pair(self, level_x, level_y, x=0, y=0):
        """
        Return approximate (negative-energy) eigenmode pair (a,b) on even/odd sublattices that arise from inserting
        a unit signal at site (x, y) into the inverse wavelet transforms at the given levels (level_x, level_y = 1, 2, ...).
        """
        assert level_x >= 1 and level_y >= 1
        a_x, b_x = self.mera1d.eigenmode_pair(level_x, x)
        a_y, b_y = self.mera1d.eigenmode_pair(level_y, y)
        assert np.allclose(a_x.range, b_x.range)
        assert np.allclose(a_y.range, b_y.range)

        n = a_x.range
        m = a_y.range
        a = np.outer(a_x.data, a_y.data)
        b = np.outer(b_x.data, b_y.data)
        return n, m, a, b

    @staticmethod
    def energy_of_mode_pair(n, m, a, b):
        """Compute energy of given single-particle mode pair (no need to go to original lattice)."""
        a = a / np.sqrt(2)
        b = b / np.sqrt(2)

        E = np.tensordot(a.conj(), b) + np.tensordot(
            b.conj()[:-1, :-1], a[1:, 1:]) - np.tensordot(
                a.conj()[1:, :], b[:-1, :]) - np.tensordot(
                    b.conj()[:, :-1], a[:, 1:])
        return -2 * np.real(E)

    def energy(self, levels_x, levels_y):
        """Compute energy of approximate ground state with branching MERA truncated at given numbers of layers."""
        E = []
        for level_x in range(1, levels_x + 1):
            for level_y in range(1, levels_y + 1):
                n, m, a, b = self.eigenmode_pair(level_x, level_y)
                e = mera2d.energy_of_mode_pair(n, m, a, b)
                E.append(e / 2**(level_x + level_y + 1))
        return np.sum(E)



import numpy as np
from .utils import *

__all__ = ['signal']


class signal:
    """A discrete signal self[n] with finite support."""

    def __init__(self, data=None, start=0):
        #: First index where signal is defined.
        self.start = start

        #: Signal data.
        self.data = np.array(data if data is not None else [])

    @property
    def stop(self):
        """Index after last index where the signal is defined."""
        return self.start + len(self.data)

    @property
    def range(self):
        """Indices where the signal is defined (and hence potentially nonzero)."""
        return np.arange(self.start, self.stop)

    def __getitem__(self, n):
        """Return element self[n]."""
        if self.start <= n < self.stop:
            return self.data[n - self.start]
        return 0

    def __repr__(self):
        return 'signal(%r, start=%d)' % (self.data, self.start)

    # this is dangerous to have (e.g. signal(other_signal) forgets the start index)
    #
    # def __array__(self):
    #     """Convert to numpy array."""
    #     return np.array(self.data)

    def to_pandas(self):
        from pandas import Series
        return Series(data=self.data, index=self.range)

    def conj(self):
        """Return complex conjugate of signal."""
        return signal(self.data.conj(), self.start)

    def norm(self):
        """Return l^2 norm of signal."""
        return np.linalg.norm(self.data)

    def isclose(self, other, **kwargs):
        """Determine whether signals are close. All keyword arguments are forwarded to numpy.allclose."""
        start, stop, a, b = self._union_align(other)
        return np.allclose(a, b, **kwargs)

    def __add__(self, other):
        """Add two signals."""
        start, stop, a, b = self._union_align(other)
        return signal(a + b, start)

    def __sub__(self, other):
        """Subtract two symbols."""
        start, stop, a, b = self._union_align(other)
        return signal(a - b, start)

    def __mul__(self, other):
        """Scalar multiplication."""
        assert np.isscalar(other)
        return signal(self.data * other, self.start)

    def __rmul__(self, other):
        """Scalar multiplication."""
        assert np.isscalar(other)
        return signal(other * self.data, self.start)

    def __truediv__(self, other):
        """Scalar division."""
        assert np.isscalar(other)
        return signal(self.data / other, self.start)

    def __neg__(self):
        """Unary negation."""
        return signal(-self.data, self.start)

    def vdot(self, other):
        """Hermitian dot product (anti-linear in first argument)."""
        start, stop, a, b = self._intersect_align(other)
        if stop <= start:
            return 0
        return np.vdot(a, b)

    def shift(self, k):
        """Return signal shifted to the right (i.e., s.start = self.start + k)."""
        return signal(self.data, self.start + k)

    def modulate(self, z):
        """Return modulated signal s[n] = self[n] * z^n (z should be complex if n can be negative)."""
        data = np.multiply(self.data, z**np.array(self.range))
        return signal(data, self.start)

    def reverse(self):
        """Return reversed signal s[n] = self[-n]."""
        return signal(self.data[::-1], -self.start - len(self.data) + 1)

    def downsample(self, repeat=1):
        """Return downsampled signal s[n] = self[2n]."""
        start = self.start
        data = self.data
        for _ in range(repeat):
            data = data[start % 2::2]
            start = (start + 1) // 2
        return signal(data, start)

    def upsample(self):
        """Return upsampled signal, s[2n] = self[n]."""
        if self.data.size == 0:
            return signal()
        start = self.start * 2
        data = np.zeros(2 * self.data.size - 1, dtype=self.data.dtype)
        data[::2] = self.data
        return signal(data, start)

    def convolve(self, other):
        """Return convolution of self and other."""
        if self.data.size == 0 or other.data.size == 0:
            return signal()
        data = np.convolve(self.data, other.data)
        start = self.start + other.start
        return signal(data, start)

    def ft(self, omega):
        """Return periodic Fourier transform (see utils.dtft)."""
        return dtft(self.range, self.data, omega)

    def _intersect_align(self, other):
        start = max(self.start, other.start)
        stop = min(self.stop, other.stop)

        # truncate
        a = self.data[start - self.start:stop - self.start]
        b = other.data[start - other.start:stop - other.start]
        return start, stop, a, b

    def _union_align(self, other):
        start = min(self.start, other.start)
        stop = max(self.stop, other.stop)

        # pad by zeros
        a = np.zeros(stop - start, dtype=self.data.dtype)
        b = np.zeros(stop - start, dtype=other.data.dtype)
        a[self.start - start:self.stop - start] = self.data
        b[other.start - start:other.stop - start] = other.data

        return start, stop, a, b



from .signal import *

__all__ = ['orthogonal_wavelet', 'DAUBECHIES_D4']


class orthogonal_wavelet:
    """Orthogonal wavelet consisting of a scaling (low-pass) filter and a wavelet (high-pass) filter that together form a conjugate mirror filter (CMF) pair."""

    def __init__(self, scaling_filter, wavelet_filter):
        #: Scaling filter (low-pass filter).
        self.scaling_filter = scaling_filter

        #: Wavelet filter (high-pass filter).
        self.wavelet_filter = wavelet_filter

    @staticmethod
    def from_scaling_filter(scaling_filter):
        """Construct orthogonal wavelet from scaling filter."""
        # could also use scipy.signal.qmf() and translate manually
        wavelet_filter = -scaling_filter.conj().modulate(-1.).shift(
            -1).reverse()
        return orthogonal_wavelet(scaling_filter, wavelet_filter)

    @staticmethod
    def from_wavelet_filter(wavelet_filter):
        """Construct orthogonal wavelet from wavelet filter."""
        scaling_filter = -wavelet_filter.reverse().shift(1).modulate(
            -1.).conj()
        return orthogonal_wavelet(scaling_filter, wavelet_filter)

    def analyze(self, s):
        """Decompose signal into scaling and wavelet coefficients."""
        scaling = s.convolve(self.scaling_filter.reverse()).downsample()
        wavelet = s.convolve(self.wavelet_filter.reverse()).downsample()
        return (scaling, wavelet)

    def reconstruct(self, scaling=None, wavelet=None):
        """Reconstruct signal from scaling and wavelet coefficients."""
        if scaling is None:
            scaling = signal()
        if wavelet is None:
            wavelet = signal()
        return (scaling.upsample().convolve(self.scaling_filter) +
                wavelet.upsample().convolve(self.wavelet_filter))

    def scaling_function(self, L):
        """Return scaling function at dyadic approximation 2^{-L}."""
        s = self._casade(L, scaling=signal([1]))
        return s.range * 2**-L, s.data * 2**(L / 2)

    def wavelet_function(self, L):
        """Return wavelet function at dyadic approximation 2^{-L}."""
        s = self._casade(L, wavelet=signal([1]))
        return s.range * 2**-L, s.data * 2**(L / 2)

    def _casade(self, L, wavelet=None, scaling=None):
        """Starting from scaling and wavelet coefficients at level L, return output of inverse wavelet transform."""
        # there is some numerical instability in scipy's implementation of the cascade algorithm (as compared to our code
        # below and to Matlab's wavefun); otherwise we could simply use scipy.signal.cascade(self.scaling_filter.data, L)
        s = self.reconstruct(scaling=scaling, wavelet=wavelet)
        for l in range(L - 1):
            s = self.reconstruct(scaling=s)
        return s


DAUBECHIES_D4_SCALING_FILTER = signal([
    0.482962913145,
    0.836516303738,
    0.224143868042,
    -0.129409522551,
])

DAUBECHIES_D4 = orthogonal_wavelet.from_scaling_filter(
    DAUBECHIES_D4_SCALING_FILTER)



import numpy as np
import scipy.signal, scipy.linalg
from .utils import *
from .signal import *
from .wavelets import *

__all__ = [
    'allpass', 'leja', 'sfact', 'selesnick_hwlet', 'evenbly_white_hwlet'
]


def allpass(tau, L):
    """
    Return the filter d[n] such that

      A(z) = z^{-L} D(1/z) / D(z)

    approximates A(z) = z^{-tau}.

    The length of the filter d[n] is L+1.
    """
    n = np.arange(L)
    x = np.r_[1, (L - n) * (L - n - tau) / (n + 1) / (n + 1 + tau)]
    return np.cumprod(x)


def leja(a):
    """
    Leja ordering of given numbers:

    * |a[0]| = max_j |a[j]|
    * prod_{i=0}^{k-1} |a[k] - a[i]| = max_j prod_{i=0}^{k-1} |a[j] - a[i]|

    When used as a preprocessing for np.poly it increases its numerical precision.
    """
    n = len(a)
    c = np.argmax(np.abs(a))
    a[[c, 0]] = a[[0, c]]
    for k in range(1, n):
        A = np.abs(
            a[:, np.newaxis][:, [0] * k] - a[np.newaxis, :k][[0] * n, :])
        A = np.prod(A, -1)
        c = np.argmax(A)
        a[[k, c]] = a[[c, k]]
    return a


def sfact(h, min_phase=False, eps=1e-5):
    """
    Return a mid-phase (or min-phase) spectral factorization of the polynomial h of degree 2n; i.e., a polynomial g of degree n such that

      h(X) = X^n g(X) g(1/X)

    The min_phase parameter is ignored if h is a complex signal.
    This code is inspired by Selesnick's sfactM.m and sfact.m.
    """
    assert len(h) % 2 == 1, 'Polynomial should have even degree.'
    isreal = np.all(np.isreal(h))

    # find roots of original polynomials
    roots = np.roots(h)

    # classify roots on unit circle
    roots_circ = roots[np.abs(np.abs(roots) - 1) < eps]
    if min_phase and roots_circ:
        raise NotImplementedError

    plus_one = np.abs(roots_circ - 1) < eps
    minus_one = np.abs(roots_circ + 1) < eps
    others = ~(plus_one | minus_one)
    pos_angle = (np.angle(roots_circ) > 0) & others
    neg_angle = (np.angle(roots_circ) < 0) & others

    num_plus_one = sum(plus_one)
    num_minus_one = sum(minus_one)
    num_pos_angle = sum(pos_angle)
    num_neg_angle = sum(neg_angle)
    assert num_plus_one % 2 == 0, 'The root +1 should appear an even number of times.'
    assert num_minus_one % 2 == 0, 'The root -1 should appear an even number of times.'
    assert num_pos_angle == num_neg_angle, 'Should have as many eigenvalues e^{i phi} as e^{-i phi}.'
    assert num_plus_one + num_minus_one + num_pos_angle + num_neg_angle == len(
        roots_circ)

    # collect half the +1's, half the -1's, and the intersperse positive with negative ones (to increase numerical stability)
    roots_pos_angle = roots_circ[pos_angle]
    roots_circ = np.r_[roots_pos_angle[::2], 1 / roots_pos_angle[1::2], [+1] *
                       (num_plus_one // 2), [-1] * (num_minus_one // 2)]

    # roots inside unit disk
    roots_int = roots[np.abs(roots) <= 1 - eps]
    if isreal and not min_phase:
        pos_imags, reals = scipy.signal.filter_design._cplxreal(roots_int)
        A1 = np.r_[pos_imags[::2], pos_imags[::2].conj()]
        A2 = np.r_[pos_imags[1::2], pos_imags[1::2].conj()]
        imags = np.r_[1 / A1, A2]
        reals = np.r_[1 / reals[::2], reals[1::2]]
        roots_int = np.r_[imags, reals]

    # roots of the spectral factorization
    roots = np.r_[roots_circ, roots_int]
    roots = leja(roots)

    # build corresponding polynomial
    g = np.poly(roots)
    if isreal:
        g = np.real(g)
    g = g * np.sqrt(h[-1] / (g[0] * g[-1]))
    if min(g) + max(g) < 0:
        g = -g

    # check that g is indeed a spectral factor of h
    assert signal(h).isclose(signal(np.convolve(g, g[::-1])))
    return g


def selesnick_hwlet(K, L, min_phase=False):
    """
    Return Selesnick's Hilbert transform wavelet pair (h, g).

    The parameter K determines the number of zeros at z=-1.
    The parameter L determines the support of the filter implementing the fractional delay.

    The length of both scaling filters is 2(K+L).
    This code is inspired by Selesnick's hwlet.m.
    """
    d = allpass(1 / 2, L)

    # filter for z^(K+L) S(z)
    s1 = scipy.special.binom(2 * K, np.arange(2 * K + 1))
    s2 = np.convolve(d, d[::-1])
    s = np.convolve(s1, s2)

    # solve convolution system for z^(K+L-1) R(z)
    A = convmtx(s, 2 * (K + L) - 1)
    A = A[1::2]
    b = np.zeros(2 * (K + L) - 1)
    b[K + L - 1] = 1
    r = np.linalg.solve(A, b)

    # find spectral factor Q(z) and compute filter for z^K F(z)
    q = sfact(r, min_phase=min_phase)
    b = scipy.special.binom(K, np.arange(K + 1))
    f = np.convolve(q, b)
    h = np.convolve(f, d)
    g = np.convolve(f, d[::-1])

    # build orthogonal wavelet
    h = orthogonal_wavelet.from_scaling_filter(signal(h))
    g = orthogonal_wavelet.from_scaling_filter(signal(g))
    return h, g


def evenbly_white_hwlet():
    """
    Return Evenbly-White's filter pair of length 4.
    """
    h_s = signal(
        np.array([-0.12940952, 0.22414387, 0.8365163, 0.48296291]), start=-2)
    g_s = signal(
        np.array([0.48296291, 0.8365163, 0.22414387, -0.12940952]), start=0)
    h = orthogonal_wavelet.from_scaling_filter(h_s)
    g = orthogonal_wavelet.from_scaling_filter(g_s)
    return h, g



import numpy as np
import scipy.fftpack

__all__ = ['convmtx', 'ctft', 'dtft', 'dtft2d']


def convmtx(h, N):
    """Return convolution matrix for kernel h and input signals of length N."""
    return scipy.linalg.toeplitz(np.r_[h, [0] * (N - 1)], np.zeros(N))


def ctft(x, f, domega=0.01, omega_max=None):
    """
    Approximate Fourier transform of a compactly-supported continuous signal.

    We assume that the samples x[n] are equally spaced.
    """
    dx = x[1] - x[0]
    if omega_max is None:
        # use Nyquist frequency as cut-off
        omega_max = 1 / (2 * dx)
    else:
        # downsample (this is an optimization)
        target_dx = 1 / (2 * omega_max)
        if target_dx >= 2 * dx:
            s = int(target_dx / dx)
            x = x[::s]
            f = f[::s]
            dx = s * dx

    # pad by zeros so that we obtain the desired frequency resolution
    target_size = max(int(1 / (domega * dx) + 1), f.size)
    pad_left = (target_size - f.size) // 2
    pad_right = target_size - f.size - pad_left
    x = np.r_[x[0] + np.arange(-pad_left, 0) * dx, x,
              x[-1] + np.arange(1, pad_right + 1) * dx]
    f = np.r_[np.zeros(pad_left), f, np.zeros(pad_right)]

    # approximate the continuous-time Fourier transform
    omega = np.fft.fftfreq(f.size, dx) * 2 * np.pi
    g = np.fft.fft(f)
    g *= dx * np.exp(-1j * omega * x[0]) / np.sqrt(2 * np.pi)

    # reshuffle so that the omega-coordinates are contiguous
    half = (f.size + 1) // 2
    omega = np.r_[omega[half:], omega[:half]]
    g = np.r_[g[half:], g[:half]]

    mask = (-omega_max <= omega) & (omega <= omega_max)
    return (omega[mask], g[mask])


def dtft(n, s, omega):
    """Periodic Fourier transform of a discrete signal s[n]."""
    return np.sum(
        s[:, np.newaxis] * np.exp(
            -1j * n[:, np.newaxis] * omega[np.newaxis, :]),
        axis=0)


def dtft2d(n, m, s, omega):
    """Periodic 2D Fourier transform of a discrete signal s[n,m]."""
    # sample discrete-time Fourier transform
    f_half = np.sum(
        s[:, :, np.newaxis] * np.exp(-1j * m[np.newaxis, :, np.newaxis] *
                                     omega[np.newaxis, np.newaxis, :]),
        axis=1)
    f = np.sum(
        f_half[:, np.newaxis, :] * np.exp(-1j * n[:, np.newaxis, np.newaxis] *
                                          omega[np.newaxis, :, np.newaxis]),
        axis=0)
    return f



import ast, re, io, os.path
from setuptools import setup

# determine version (adapted from mitsuhiko)
VERSION_RE = re.compile(r'__version__\s+=\s+(.*)')
with open('pyfermions/__init__.py', 'rb') as f:
    version = VERSION_RE.search(f.read().decode('utf-8')).group(1)
    version = str(ast.literal_eval(version))

# read long description and convert to RST
long_description = io.open(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), 'README.md'),
    encoding='utf-8').read()

setup(
    name='pyfermions',
    version=version,
    description=
    'Rigorous free fermion entanglement renormalization from wavelet theory',
    long_description=long_description,
    long_description_content_type='text/markdown',
    maintainer='Michael Walter',
    maintainer_email='michael.walter@stanford.edu',
    url='https://github.com/catch22/pyfermions',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
    packages=['pyfermions'],
    install_requires=[
        'matplotlib',
        'numpy',
        'scipy',
        'pandas',
        'jupyter',
    ],
    extras_require={
        'dev': [
            'pytest',
            'wheel',
            'yapf',
            'twine',
        ],
    })



import numpy as np
from .signal import *
from .wavelets import *


def random_signal():
    N = 100
    return signal(
        np.random.rand(N) + 1j * np.random.rand(N), np.random.randint(-N, N))


def test_analyze_and_reconstruct():
    a = random_signal()
    assert DAUBECHIES_D4.reconstruct(*DAUBECHIES_D4.analyze(a)).isclose(a)


def test_reconstruct_and_analyze():
    a = random_signal()
    b = random_signal()
    A, B = DAUBECHIES_D4.analyze(DAUBECHIES_D4.reconstruct(a, b))
    assert a.isclose(A)
    assert b.isclose(B)


def test_scaling_from_reconstruct():
    scaling_filter = DAUBECHIES_D4.reconstruct(signal([1]), signal())
    assert scaling_filter.isclose(DAUBECHIES_D4.scaling_filter)


def test_from_scaling_filter():
    assert orthogonal_wavelet.from_scaling_filter(
        DAUBECHIES_D4.scaling_filter).wavelet_filter.isclose(
            DAUBECHIES_D4.wavelet_filter)


def test_from_wavelet_filter():
    assert orthogonal_wavelet.from_wavelet_filter(
        DAUBECHIES_D4.wavelet_filter).scaling_filter.isclose(
            DAUBECHIES_D4.scaling_filter)
